from platon import Web3
from platon.datastructures import AttributeDict

from platon_aide.economic import gas, Economic
from platon_aide.module import Module
from platon_aide.staking import Staking
from platon_aide.utils import contract_transaction


class _DelegateInfo(AttributeDict):
    """ 委托信息的属性字典类
    """
    Addr: str
    NodeId: str
    StakingBlockNum: int
    DelegateEpoch: int
    Released: int
    ReleasedHes: int
    RestrictingPlan: int
    RestrictingPlanHes: int
    CumulativeIncome: int


class Delegate(Module):

    def __init__(self, web3: Web3):
        super().__init__(web3)
        self._module_type = 'inner-contract'
        self._result_type = 'event'
        self._get_node_info()
        self._economic = Economic(web3)

    @property
    def _staking_block_number(self):
        staking = Staking(self.web3)
        return staking.staking_info.StakingBlockNum

    @contract_transaction
    def delegate(self,
                 amount=None,
                 balance_type=0,
                 node_id=None,
                 txn=None,
                 private_key=None,
                 ):
        """ 委托节点，以获取节点的奖励分红
        """
        amount = amount or self._economic.add_staking_limit
        node_id = node_id or self._node_id
        return self.web3.ppos.delegate.delegate(node_id, balance_type, amount)

    @contract_transaction
    def withdrew_delegate(self,
                          amount=0,
                          staking_block_identifier=None,
                          node_id=None,
                          txn=None,
                          private_key=None,
                          ):
        """
        撤回对节点的委托，可以撤回部分委托
        注意：因为节点可能进行过多次质押/撤销质押，会使得委托信息遗留，因此撤回委托时必须指定节点质押区块
        """
        node_id = node_id or self._node_id
        amount = amount or self._economic.add_staking_limit
        staking_block_identifier = staking_block_identifier or self._staking_block_number

        return self.web3.ppos.delegate.withdrew_delegate(node_id,
                                                         staking_block_identifier,
                                                         amount,
                                                         )

    def get_delegate_info(self,
                          address=None,
                          node_id=None,
                          staking_block_identifier=None,
                          ):
        """ 获取地址对某个节点的某次质押的委托信息
        注意：因为节点可能进行过多次质押/撤销质押，会使得委托信息遗留，因此获取委托信息时必须指定节点质押区块
        """
        if self.default_account:
            address = address or self.default_account.address
        node_id = node_id or self._node_id
        staking_block_identifier = staking_block_identifier or self._staking_block_number

        delegate_info = self.web3.ppos.delegate.get_delegate_info(address, node_id, staking_block_identifier)
        if delegate_info == 'Query delegate info failed:Delegate info is not found':
            return None
        else:
            return _DelegateInfo(delegate_info)

    def get_delegate_list(self, address=None):
        """ 获取地址的全部委托信息
        """
        if self.default_account:
            address = address or self.default_account.address
        # return self.web3.ppos.delegate.get_delegate_list(address)
        delegate_list = self.web3.ppos.delegate.get_delegate_list(address)
        if delegate_list == 'Retreiving delegation related mapping failed:RelatedList info is not found':
            return None
        else:
            return [_DelegateInfo(delegate_info) for delegate_info in delegate_list]

    @contract_transaction
    def withdraw_delegate_reward(self,
                                 txn=None,
                                 private_key=None,
                                 ):
        """ 提取委托奖励，会提取委托了的所有节点的委托奖励
        """
        return self.web3.ppos.delegate.withdraw_delegate_reward()

    def get_delegate_reward(self,
                            address=None,
                            node_ids=None
                            ):
        """ 获取委托奖励信息，可以根据节点id过滤
        """
        if self.default_account:
            address = address or self.default_account.address
        node_ids = node_ids or []
        return self.web3.ppos.delegate.get_delegate_reward(address, node_ids)



from time import time

from platon import Web3
from platon.datastructures import AttributeDict

from platon_aide.module import Module
from platon_aide.utils import contract_transaction


class _ChainVersion(AttributeDict):
    integer: int
    major: int
    minor: int
    patch: int


class _BaseProposal(AttributeDict):
    ProposalID: str
    Proposer: str
    ProposalType: int
    PIPID: str
    SubmitBlock: int
    EndVotingBlock: int


class _VersionProposal(_BaseProposal):
    EndVotingRounds: int
    ActiveBlock: int
    NewVersion: int


class _ParamProposal(_BaseProposal):
    Module: str
    Name: str
    NewVersion: str


class _CancelProposal(_BaseProposal):
    EndVotingRounds: str
    EndVotingBlock: str
    TobeCanceled: str


class _TextProposal(_BaseProposal):
    pass


class _ProposalVotes(AttributeDict):
    accuVerifiers: int
    yeas: int
    nays: int
    abstentions: int
    particiRatio: float
    yeasRatio: float


class _ProposalResult(_ProposalVotes):
    proposalID: str
    status: int
    canceledBy: str


second = lambda: str(int(time()))


def to_attribute_proposal(proposal):
    _type = proposal['ProposalType']
    wizard = {
        1: _TextProposal,
        2: _VersionProposal,
        3: _ParamProposal,
        4: _CancelProposal,
    }
    attribute_proposal = wizard[_type]
    return attribute_proposal(proposal)


class Govern(Module):

    def __init__(self, web3: Web3):
        super().__init__(web3)
        self._module_type = 'inner-contract'
        self._result_type = 'event'
        self._get_node_info()

    @property
    def chain_version(self):
        version = self.web3.pip.get_chain_version()
        version_byte = int(version).to_bytes(length=3, byteorder='big')
        return _ChainVersion({
            'integer': version,
            'major': version_byte[0],
            'minor': version_byte[1],
            'patch': version_byte[2],
        })

    @contract_transaction
    def version_proposal(self,
                         version,
                         voting_rounds=4,
                         pip_number=second(),
                         node_id=None,
                         txn=None,
                         private_key=None,
                         ):
        """ 提交版本提案，实现链上共识硬分叉版本的升级
        """
        node_id = node_id or self._node_id
        return self.web3.pip.submit_version_proposal(node_id, pip_number, version, voting_rounds)

    @contract_transaction
    def param_proposal(self,
                      module,
                      name,
                      value,
                      pip_number=second(),
                      node_id=None,
                      txn=None,
                      private_key=None,
                      ):
        """ 提交参数提案，修改链上可治理参数
        """
        node_id = node_id or self._node_id
        return self.web3.pip.submit_param_proposal(node_id, pip_number, module, name, value)

    @contract_transaction
    def cancel_proposal(self,
                        proposal_id,
                        voting_rounds=4,
                        node_id=None,
                        pip_number=second(),
                        txn=None,
                        private_key=None,
                        ):
        """ 提交取消提案
        """
        node_id = node_id or self._node_id
        return self.web3.pip.submit_cancel_proposal(node_id, pip_number, voting_rounds, proposal_id)

    @contract_transaction
    def text_proposal(self,
                      pip_number=second(),
                      node_id=None,
                      txn=None,
                      private_key=None,
                      ):
        """ 提交文本提案，文本提案不对链上产生影响，仅做pip投票意见收集作用
        """
        node_id = node_id or self._node_id
        return self.web3.pip.submit_text_proposal(node_id, pip_number)

    @contract_transaction
    def vote(self,
             proposal_id,
             option,
             node_id=None,
             version=None,
             version_sign=None,
             txn=None,
             private_key=None,
             ):
        """ 对提案进行投票
        """
        node_id = node_id or self._node_id
        version = version or self._version
        version_sign = version_sign or self._version_sign
        return self.web3.pip.vote(node_id, proposal_id, option, version, version_sign)

    @contract_transaction
    def declare_version(self,
                        node_id=None,
                        version=None,
                        version_sign=None,
                        txn=None,
                        private_key=None,
                        ):
        """ 向链上声明节点版本，以获得参与共识出块的资格
        """
        node_id = node_id or self._node_id
        version = version or self._version
        version_sign = version_sign or self._version_sign
        return self.web3.pip.declare_version(node_id, version, version_sign)

    def get_proposal(self, proposal_id):
        """ 获取提案信息
        """
        proposal = self.web3.pip.get_proposal(proposal_id)
        return to_attribute_proposal(proposal)

    def get_proposal_result(self, proposal_id):
        """ 获取提案投票结果信息
        """
        proposal_result = self.web3.pip.get_proposal_result(proposal_id)
        if not proposal_result:
            raise ValueError('proposal is not found.')
        if type(proposal_result) is str:
            raise ValueError(f'{proposal_result}.')

        partici_count = proposal_result['yeas'] + proposal_result['nays'] + proposal_result['abstentions']
        proposal_result['particiRatio'] = partici_count / proposal_result['accuVerifiers']
        proposal_result['yeasRatio'] = proposal_result['yeas'] / partici_count if partici_count else 0

        return _ProposalResult(proposal_result)

    def get_proposal_votes(self, proposal_id, block_identifier='latest'):
        """ 获取提案实时投票信息
        """
        proposal_votes = self.web3.pip.get_proposal_votes(proposal_id, block_identifier)
        if not proposal_votes:
            raise ValueError('proposal is not found.')

        partici_count = proposal_votes[1] + proposal_votes[2] + proposal_votes[3]

        return _ProposalVotes({
            'accuVerifiers': proposal_votes[0],
            'yeas': proposal_votes[1],
            'nays': proposal_votes[2],
            'abstentions': proposal_votes[3],
            'particiRatio': partici_count / proposal_votes[0],
            'yeasRatio': proposal_votes[1] / partici_count if partici_count else 0,
        })

    def proposal_list(self, proposal_type=None):
        """ 获取提案列表，可以根据提案类型过滤
        """
        proposal_list = self.web3.pip.proposal_list()
        if type(proposal_list) is not list:
            raise ValueError('proposals is not found.')

        return [to_attribute_proposal(proposal) for proposal in proposal_list]

    def get_govern_param(self, module, name):
        """ 获取可治理参数的值
        """
        return self.web3.pip.get_govern_param(module, name)

    def govern_param_list(self, module=None):
        """ 获取可治理参数列表信息
        """
        return self.web3.pip.govern_param_list(module)



from platon import Web3
from gql import gql

from platon_aide.module import Module
from platon_aide.utils import get_gql


class Graphql(Module):

    def __init__(self, uri: str, web3: Web3 = None):
        super().__init__(web3)
        self._gql = get_gql(uri)

    def execute(self, content):
        return self._gql.execute(gql(content))



from platon import Web3

from platon_aide.module import Module
from platon_aide.utils import contract_transaction


class Slashing(Module):

    def __init__(self, web3: Web3):
        super().__init__(web3)
        self._module_type = 'inner-contract'
        self._result_type = 'event'
        self._get_node_info()

    @contract_transaction
    def report_duplicate_sign(self,
                              report_type,
                              data,
                              txn=None,
                              private_key=None,
                              ):
        return self.web3.ppos.slashing.report_duplicate_sign(report_type, data)

    def check_duplicate_sign(self,
                             report_type,
                             block_identifier,
                             node_id=None,
                             ):
        node_id = node_id or self._node_id
        return self.web3.ppos.slashing.check_duplicate_sign(report_type, node_id, block_identifier)



from platon import Web3
from platon.datastructures import AttributeDict

from platon_aide.economic import Economic
from platon_aide.module import Module
from platon_aide.utils import contract_transaction


class _StakingInfo(AttributeDict):
    """ 质押信息的属性字典类
    """
    NodeId: str
    StakingAddress: str
    BenefitAddress: str
    RewardPer: int
    NextRewardPer: int
    StakingTxIndex: int
    ProgramVersion: int
    Status: int
    StakingEpoch: int
    StakingBlockNum: int
    Shares: int
    Released: int
    ReleasedHes: int
    RestrictingPlan: int
    RestrictingPlanHes: int
    ExternalId: int
    NodeName: str
    Website: str
    Details: str
    DelegateEpoch: int
    DelegateTotal: int
    DelegateTotalHes: int
    DelegateRewardTotal: int


class Staking(Module):
    _node_id: str
    _bls_pubkey: str
    _bls_proof: str
    _version: str
    _version_sign: str

    def __init__(self, web3: Web3):
        super().__init__(web3)
        self._module_type = 'inner-contract'
        self._result_type = 'event'
        self._get_node_info()
        self._economic = Economic(web3)

    @property
    def staking_info(self):
        candidate_info = self.get_candidate_info(self._node_id)
        if candidate_info:
            return candidate_info
        raise AttributeError('the node has no staking information.')

    @contract_transaction
    def create_staking(self,
                       amount=None,
                       balance_type=0,
                       node_id=None,
                       benefit_address=None,
                       node_name='',
                       external_id='',
                       details='',
                       website='',
                       reward_per=0,
                       version=None,
                       version_sign=None,
                       bls_pubkey=None,
                       bls_proof=None,
                       txn=None,
                       private_key=None,
                       ):

        if not benefit_address:
            benefit_account = self.web3.platon.account.from_key(private_key) if private_key else self.default_account
            if not benefit_account:
                raise ValueError('the benefit address cannot be empty')
            benefit_address = benefit_account.address

        node_id = node_id or self._node_id
        amount = amount or self._economic.staking_limit
        version = version or self._version
        version_sign = version_sign or self._version_sign
        bls_pubkey = bls_pubkey or self._bls_pubkey
        bls_proof = bls_proof or self._bls_proof
        return self.web3.ppos.staking.create_staking(balance_type, benefit_address, node_id, external_id,
                                                     node_name, website, details, amount, reward_per,
                                                     version, version_sign, bls_pubkey, bls_proof
                                                     )

    @contract_transaction
    def increase_staking(self,
                         balance_type=0,
                         node_id=None,
                         amount=None,
                         txn=None,
                         private_key=None,
                         ):
        node_id = node_id or self._node_id
        amount = amount or self._economic.add_staking_limit
        return self.web3.ppos.staking.increase_staking(node_id, balance_type, amount)

    @contract_transaction
    def withdrew_staking(self, node_id=None, txn=None, private_key=None):
        node_id = node_id or self._node_id
        return self.web3.ppos.staking.withdrew_staking(node_id)

    @contract_transaction
    def edit_candidate(self,
                       benifit_address=None,
                       node_id=None,
                       reward_per=None,
                       external_id=None,
                       node_name=None,
                       website=None,
                       details=None,
                       txn=None,
                       private_key=None,
                       ):
        node_id = node_id or self._node_id
        return self.web3.ppos.staking.edit_staking(node_id, benifit_address, reward_per, external_id,
                                                   node_name, website, details
                                                   )

    def get_verifier_list(self):
        verifier_list = self.web3.ppos.staking.get_verifier_list()
        return [_StakingInfo(verifier) for verifier in verifier_list]

    def get_validator_list(self):
        validator_list = self.web3.ppos.staking.get_validator_list()
        return [_StakingInfo(validator) for validator in validator_list]

    def get_candidate_list(self):
        candidate_list = self.web3.ppos.staking.get_candidate_list()
        return [_StakingInfo(candidate) for candidate in candidate_list]

    def get_candidate_info(self, node_id=None):
        node_id = node_id or self._node_id
        staking_info = self.web3.ppos.staking.get_candidate_info(node_id)
        if staking_info == 'Query candidate info failed:Candidate info is not found':
            return None
        else:
            return _StakingInfo(staking_info)

    def get_block_reward(self):
        return self.web3.ppos.staking.get_block_reward()

    def get_staking_reward(self):
        return self.web3.ppos.staking.get_staking_reward()

    def get_avg_block_time(self):
        return self.web3.ppos.staking.get_avg_block_time()



import json
from decimal import Decimal
from typing import Literal

from platon import Web3
from platon.datastructures import AttributeDict

from platon_aide.module import Module
from platon_aide.utils import ec_recover


class _ChainData(AttributeDict):
    maxEpochMinutes: int
    nodeBlockTimeWindow: int
    perRoundBlocks: int
    maxConsensusVals: int
    additionalCycleTime: int


class _StakingData(AttributeDict):
    stakeThreshold: int
    operatingThreshold: int
    maxValidators: int
    unStakeFreezeDuration: int
    rewardPerMaxChangeRange: int
    rewardPerChangeInterval: int


class _SlashingData(AttributeDict):
    slashFractionDuplicateSign: int
    duplicateSignReportReward: int
    maxEvidenceAge: int
    slashBlocksReward: int
    zeroProduceCumulativeTime: int
    zeroProduceNumberThreshold: int
    zeroProduceFreezeDuration: int


class _GovernData(AttributeDict):
    versionProposalVoteDurationSeconds: int
    versionProposalSupportRate: int
    textProposalVoteDurationSeconds: int
    textProposalVoteRate: int
    textProposalSupportRate: int
    cancelProposalVoteRate: int
    cancelProposalSupportRate: int
    paramProposalVoteDurationSeconds: int
    paramProposalVoteRate: int
    paramProposalSupportRate: int


class _RewardData(AttributeDict):
    newBlockRate: int
    platonFoundationYear: int
    increaseIssuanceRatio: int
    TheNumberOfDelegationsReward: int


class _RestrictingData(AttributeDict):
    minimumRelease: int


class _GasData:
    transferGas: int = 21000
    restrictingGas: int = 100000
    governGasPrice: int = 2100000000000000


class GenesisData:
    """ PlatON创世数据
    """

    def __init__(self, web3: Web3):
        economic_config = json.loads(web3.debug.economic_config())
        self.chain = _ChainData(economic_config['common'])
        self.restricting = _RestrictingData(economic_config.get('restricting')) if economic_config.get('restricting') else None
        self.staking = _StakingData(economic_config['staking'])
        self.govern = _GovernData(economic_config['gov'])
        self.reward = _RewardData(economic_config['reward'])
        self.slashing = _SlashingData(economic_config['slashing'])
        self.gas = _GasData()


gas = _GasData()


class Economic(Module):

    def __init__(self, web3: Web3):
        super().__init__(web3)
        self.genesis = GenesisData(web3)
        self._get_node_info()

    # 区块
    @property
    def block_time(self):
        """ 区块时长/s
        """
        # todo: 获取区块平均时间
        return int(self.genesis.chain.nodeBlockTimeWindow // self.genesis.chain.perRoundBlocks)

    # 窗口期
    @property
    def round_time(self):
        """ 窗口期时长/m
        """
        return self.round_blocks * self.block_time

    @property
    def round_blocks(self):
        """ 窗口期区块数量
        """
        return self.genesis.chain.perRoundBlocks

    # 共识轮
    @property
    def consensus_time(self):
        """ 共识轮时长/m
        """
        return self.consensus_rounds * self.round_time

    @property
    def consensus_blocks(self):
        """ 共识轮区块数量
        """
        return self.consensus_rounds * self.round_blocks

    @property
    def consensus_rounds(self):
        """ 共识轮窗口期数量
        """
        return self.validator_count

    # 结算周期
    @property
    def epoch_time(self):
        """ 结算周期时长/m
        """
        return self.epoch_consensus * self.consensus_time

    @property
    def epoch_blocks(self):
        """ 结算周期区块数量
        """
        return self.epoch_consensus * self.consensus_blocks

    @property
    def epoch_rounds(self):
        """ 结算周期窗口期数量
        """
        return self.epoch_consensus * self.consensus_rounds

    @property
    def epoch_consensus(self):
        """ 结算周期共识轮数量
        """
        return (self.genesis.chain.maxEpochMinutes * 60) // self.consensus_time

    @property
    def increasing_time(self):
        """ 增发周期时长/m
        """
        return self.increasing_epoch * self.epoch_time

    @property
    def increasing_blocks(self):
        """ 增发周期区块数
        """
        return self.increasing_rounds * self.round_blocks

    @property
    def increasing_rounds(self):
        """ 增发周期窗口期数
        """
        return self.increasing_consensus * self.consensus_rounds

    @property
    def increasing_consensus(self):
        """ 增发周期共识轮数
        """
        return self.increasing_epoch * self.epoch_consensus

    @property
    def increasing_epoch(self):
        """ 增发周期结算周期数
        """
        # todo: 修改为实时计算
        return (self.genesis.chain.additionalCycleTime * 60) // self.epoch_time

    @property
    def validator_count(self):
        """ 共识验证人数量
        """
        return self.genesis.chain.maxConsensusVals

    @property
    def staking_limit(self):
        """ 质押最小金额限制
        """
        return self.genesis.staking.stakeThreshold

    @property
    def add_staking_limit(self):
        """ 增持质押最小金额限制
        """
        return self.genesis.staking.operatingThreshold

    @property
    def delegate_limit(self):
        """ 委托最小金额限制
        """
        return self.genesis.staking.operatingThreshold

    @property
    def unstaking_freeze_epochs(self):
        """ 解质押后，质押金额冻结的结算周期数
        """
        return self.genesis.staking.unStakeFreezeDuration

    @property
    def not_block_slash_rate(self):
        """ 节点零出块处罚时，罚金对应的区块奖励倍数
        """
        return self.genesis.slashing.slashBlocksReward

    def param_proposal_epochs(self):
        """ 参数提案投票期的结算周期数
        """
        return self.genesis.govern.paramProposalVoteDurationSeconds // self.epoch_time

    def text_proposal_epochs(self):
        """ 文本提案投票期的结算周期数
        """
        return self.genesis.govern.textProposalVoteDurationSeconds // self.epoch_time

    def get_verifier_count(self, epoch=None):
        """ 获取结算周期的验证人数
        注意：目前只能获取当前结算周期的验证人数，无法获取历史增发周期的验证人数
        # todo: 推动底层改进，以实现逻辑自洽
        """
        # if not epoch:
        #     epoch, _ = self.get_period_info(self.web3.platon.block_number, period_type='epoch')

        verifier_list = self.web3.ppos.staking.get_verifier_list()
        return len(verifier_list)

    def get_blocks_from_miner(self, start=None, end=None, node_id=None):
        """ 获取节点出块数
        """
        block_count = 0
        for bn in range(start, end):
            block = self.web3.platon.get_block(bn)
            public_key = ec_recover(block)
            if node_id in public_key:
                block_count = block_count + 1
        return block_count

    def get_rewards_from_epoch(self, epoch=None):
        """ 根据增发周期，获取增发周期的奖励信息
        注意：目前只能获取当前增发周期的奖励信息，无法获取历史增发周期的奖励信息
        # todo: 推动底层改进，以实现逻辑自洽
        """
        staking_reward = self.web3.ppos.staking.get_staking_reward()
        block_reward = self.web3.ppos.staking.get_block_reward()
        return staking_reward, block_reward

    @staticmethod
    def calc_staking_reward(epoch_staking_reward,
                            epoch_block_reward,
                            verifier_count,
                            block_count,
                            ):
        """
        计算一个结算周期内，节点的质押奖励

        Args:
            epoch_staking_reward: 结算周期的质押奖励，可以通过self.get_rewards_from_epoch获取
            epoch_block_reward: 结算周期的出块奖励，可以通过self.get_rewards_from_epoch获取
            verifier_count: 结算周期的验证人数
            block_count: 节点在结算周期的出块数，可以通过self.get_blocks_from_miner获取
        """
        staking_reward = Decimal(epoch_staking_reward) / Decimal(verifier_count)
        block_reward = Decimal(epoch_block_reward) * Decimal(block_count)
        return int(staking_reward), int(block_reward)

    # def __calc_staking_reward(self, node_id=None, epoch=None, verifier_count=None):
    #     """ 根据结算周期，计算节点在结算周期的质押奖励
    #     注意：因底层实现逻辑原因，本方法预期功能未能实现，请勿使用
    #     """
    #     if epoch and (not verifier_count):
    #         Warning('when passing in epoch, it is recommended to pass in the verifier count of the epoch.')
    #
    #     node_id = node_id or self._node_id
    #     if not epoch:
    #         epoch, _ = self.get_period_info(self.web3.platon.block_number, 'epoch')
    #
    #     epoch_staking_reward, epoch_block_reward = self.get_rewards_from_epoch(epoch)
    #
    #     if not verifier_count:
    #         verifier_count = self.get_verifier_count(epoch)
    #
    #     start_bn, end_bn = (epoch - 1) * self.epoch_blocks + 1, epoch * self.epoch_blocks
    #     block_count = self.get_blocks_from_miner(start_bn, end_bn, node_id=node_id)
    #
    #     return self.calc_staking_reward(epoch_staking_reward,
    #                                     epoch_block_reward,
    #                                     verifier_count,
    #                                     block_count,
    #                                     )

    @staticmethod
    def calc_delegate_reward(total_node_reward,
                             delegate_reward_ratio,
                             delegate_total_amount,
                             delegate_amount,
                             ):
        """
        计算一个结算周期内，账户在某个节点的委托奖励

        Args:
            total_node_reward: 节点在该结算周期的总质押奖励
            delegate_reward_ratio: 节点在该结算周期的委托分红比例
            delegate_total_amount: 节点在该结算周期的锁定期委托总额
            delegate_amount: 结算周期内，账户在该节点的锁定期委托总额
        """
        total_delegate_reward = Decimal(total_node_reward) * (Decimal(delegate_reward_ratio) / Decimal(10000))
        per_delegate_reward = Decimal(total_delegate_reward) / Decimal(delegate_total_amount)
        return int(per_delegate_reward * delegate_amount)

    def calc_report_multi_sign_reward(self, staking_amount):
        """ 计算举报双签的奖励
        """
        slashing_ratio = self.genesis.slashing.slashFractionDuplicateSign
        slashing_amount = Decimal(staking_amount) * (Decimal(slashing_ratio) / 10000)

        reward_ratio = self.genesis.slashing.duplicateSignReportReward
        report_reward = slashing_amount * (Decimal(reward_ratio) / 100)

        to_incentive_pool_amount = slashing_amount - report_reward
        return int(report_reward), int(to_incentive_pool_amount)

    def get_period_info(self,
                        block_number=None,
                        period_type: Literal['round', 'consensus', 'epoch', 'increasing'] = 'epoch'
                        ):
        """ 通过区块高度和周期类型，获取区块所在的周期数和周期结束块高
        """
        if period_type not in ['round', 'consensus', 'epoch', 'increasing']:
            raise ValueError('unrecognized period type.')

        blocks = {
            'round': self.round_blocks,
            'consensus': self.consensus_blocks,
            'epoch': self.epoch_blocks,
            'increasing': self.round_blocks,
        }
        period_blocks = blocks[period_type]

        if not block_number:
            block_number = self.web3.platon.block_number

        period = (block_number // period_blocks) + 1
        end_block = period * period_blocks

        return period, end_block



from platon import Web3

from platon_aide.economic import gas
from platon_aide.module import Module
from platon_aide.utils import contract_transaction


class Transfer(Module):

    # 转账交易
    # todo: 交易不支持返回类型为ic-event
    def transfer(self, to_address, amount, txn=None, private_key=None):
        base_txn = {
            "to": to_address,
            "gasPrice": self.web3.platon.gas_price,
            "gas": gas.transferGas,
            "data": '',
            "chainId": self.web3.chain_id,
            "value": amount,
        }
        if txn:
            base_txn.update(txn)
        txn = base_txn
        if self._result_type == 'txn':
            return txn
        return self.send_transaction(txn, private_key, self._result_type)

    @contract_transaction
    def restricting(self, release_address, plans, txn=None, private_key=None):
        return self.web3.restricting.create_restricting(release_address, plans)

    def get_balance(self, account, block_identifier=None):
        return self.web3.platon.get_balance(account, block_identifier)

    def get_restricting_info(self, release_address):
        return self.web3.restricting.get_restricting_info(release_address)



import functools
import rlp
from hexbytes import HexBytes

from platon import Web3, HTTPProvider, WebsocketProvider, IPCProvider
from platon._utils.inner_contract import InnerContractEvent
from platon._utils.threads import Timeout
from platon.types import BlockData
from platon_account import Account
from platon_account._utils.signing import to_standard_signature_bytes
from platon_hash.auto import keccak
from platon_keys.datatypes import Signature
from platon_typing import HexStr
from platon_utils import remove_0x_prefix, to_canonical_address

from gql import Client
from gql.transport.aiohttp import AIOHTTPTransport
from gql.transport.websockets import WebsocketsTransport


def get_web3(uri, chain_id=None, hrp=None, timeout=10):
    """ 通过rpc uri，获取web3对象。可以兼容历史platon版本
    """
    if uri.startswith('http'):
        provider = HTTPProvider
    elif uri.startswith('ws'):
        provider = WebsocketProvider
    elif uri.startswith('ipc'):
        provider = IPCProvider
    else:
        raise ValueError(f'unidentifiable uri {uri}')

    with Timeout(timeout) as t:
        while True:
            web3 = Web3(provider(uri), chain_id=chain_id, hrp=hrp)
            if web3.isConnected():
                break
            t.sleep(2)

    return web3


def get_gql(uri):
    """ 通过gql uri，获取gql对象。
    # todo: 增加超时处理
    """
    if uri.startswith('http'):
        transport = AIOHTTPTransport
    elif uri.startswith('ws'):
        transport = WebsocketsTransport
    else:
        raise ValueError(f'unidentifiable uri {uri}')

    return Client(transport=transport(uri), fetch_schema_from_transport=True)


def send_transaction(web3: Web3, txn: dict, private_key: str):
    """
    签名交易并发送，可以获取交易hash或交易回执

    Args:
        web3: Web3对象
        txn: 要发送的交易dict
        private_key: 地址私钥，用于签名交易
        returns: 指定要返回的结果，取值如下：
                - 'hash': 返回交易哈希
                - 'receipt': 返回交易回执
                - 'event': 返回内置合约的event内容
    """
    if not private_key:
        return web3.platon.send_transaction(txn)

    if not txn.get('nonce'):
        account = web3.platon.account.from_key(private_key, hrp=web3.hrp)
        txn['nonce'] = web3.platon.get_transaction_count(account.address)

    signed_txn = web3.platon.account.sign_transaction(txn, private_key, web3.hrp)
    return web3.platon.send_raw_transaction(signed_txn.rawTransaction)


def get_transaction_result(web3: Web3, tx_hash, result_type):
    """ 根据指定的result type，来获取交易的返回值
    """
    if result_type == 'hash':
        return bytes(tx_hash).hex()
    receipt = web3.platon.wait_for_transaction_receipt(tx_hash)
    if type(receipt) is bytes:
        receipt = receipt.decode('utf-8')
    if result_type == 'receipt':
        return receipt
    if result_type == 'event':
        return InnerContractEvent.get_event(receipt)


def contract_call(func):
    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):
        return func(self, *args, **kwargs).call()

    return wrapper


def contract_transaction(func):
    """ todo: 增加注释
    """

    @functools.wraps(func)
    def wrapper(self, *args, txn=None, private_key=None, **kwargs):
        # todo: 在预估gas之前，增加from地址指定
        txn = func(self, *args, private_key=private_key, **kwargs).build_transaction(txn)
        if self._result_type == 'txn':
            return txn
        return self.send_transaction(txn, private_key, self._result_type)

    return wrapper


def ec_recover(block: BlockData):
    """ 使用keccak方式，解出区块的签名节点公钥
    """
    extra = block.proofOfAuthorityData[:32]
    sign = block.proofOfAuthorityData[32:]
    raw_data = [bytes.fromhex(remove_0x_prefix(block.parentHash.hex())),
                to_canonical_address(block.miner),
                bytes.fromhex(remove_0x_prefix(block.stateRoot.hex())),
                bytes.fromhex(remove_0x_prefix(block.transactionsRoot.hex())),
                bytes.fromhex(remove_0x_prefix(block.receiptsRoot.hex())),
                bytes.fromhex(remove_0x_prefix(block.logsBloom.hex())),
                block.number,
                block.gasLimit,
                block.gasUsed,
                block.timestamp,
                extra,
                bytes.fromhex(remove_0x_prefix(block.nonce.hex()))
                ]
    hash_bytes = HexBytes(keccak(rlp.encode(raw_data)))
    signature_bytes = HexBytes(sign)
    signature_bytes_standard = to_standard_signature_bytes(signature_bytes)
    signature = Signature(signature_bytes=signature_bytes_standard)
    return remove_0x_prefix(HexStr(signature.recover_public_key_from_msg_hash(hash_bytes).to_hex()))



from setuptools import (
    setup,
    find_packages,
)


deps = {
    'platon-aide': [
        'platon.py>=1.2.0',
        'platon_account>=1.2.0',
        'platon_hash>=1.2.0',
        'platon_keys>=1.2.0',
        'platon_typing>=1.2.0',
        'platon_utils>=1.2.0',
        'rlp>=1.2.0',
        'gql>=3.0.0rc0',
    ]
}

with open('./README.md', encoding='utf-8') as readme:
    long_description = readme.read()

setup(
    name='platon-aide',
    # *IMPORTANT*: Don't manually change the version here. Use the 'bumpversion' utility.
    version='1.2.6',
    description="""An aide that helps you quickly access the platon chain and use its basic functions.""",
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Shinnng',
    author_email='shinnng@outlook.com',
    url='https://github.com/shinnng/platon-aide',
    include_package_data=True,
    install_requires=deps['platon-aide'],
    py_modules=['platon_aide'],
    extras_require=deps,
    license="MIT",
    zip_safe=False,
    package_data={'platon-aide': ['py.typed']},
    keywords='platon',
    packages=find_packages(exclude=["tests", "tests.*"]),
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
)



import warnings
from platon import Account
from platon import Web3
from platon_typing import Address

from platon_aide.utils import send_transaction, get_transaction_result


class Module:
    address: None

    def __init__(self, web3: Web3):
        self.web3 = web3
        self.default_account: Account = None
        # todo: 设置默认地址
        self.default_address: Address = None

        # 模块类型，目前仅用于判断是否可以返回event，包括：'inner-contract'
        self._module_type = ''
        # 返回结果类型，包括：txn, hash, receipt, event(仅内置合约可用)
        self._result_type = 'receipt'

    def set_default_account(self, account):
        self.default_account = account

    def _get_node_info(self):
        node_info = self.web3.node.admin.node_info()
        # self._node_id = node_info['id']                                    # todo: 增加不使用id字段的注释
        self._node_id = node_info['enode'].split('//')[1].split('@')[0]  # 请使用enode中的节点
        self._bls_pubkey = node_info['blsPubKey']
        self._bls_proof = self.web3.node.admin.get_schnorr_NIZK_prove()
        version_info = self.web3.node.admin.get_program_version()
        self._version = version_info['Version']
        self._version_sign = version_info['Sign']

    def send_transaction(self, txn, private_key, result_type=''):
        result_type = result_type or self._result_type

        if not private_key and self.default_account:
            private_key = self.default_account.privateKey.hex()[2:]

        tx_hash = send_transaction(self.web3, txn, private_key)
        return self._get_transaction_result(tx_hash, result_type)

    def _get_transaction_result(self, tx_hash, result_type):
        if result_type == 'event' and self._module_type != 'inner-contract':
            raise TypeError('result type "event" only support inner contract')

        return get_transaction_result(self.web3, tx_hash, result_type)

    def set_result_type(self, result_type):
        if result_type not in ('txn', 'hash', 'receipt', 'event'):
            raise ValueError('Unrecognized value')

        if result_type == 'event' and self._module_type != 'inner-contract':
            warnings.warn(f'result type "event" only support inner contract, '
                    f'try set {self.__class__.__name__} result type to "receipt"', RuntimeWarning)
            result_type = 'receipt'

        self._result_type = result_type



import time

from platon.middleware import gplaton_poa_middleware

from platon_aide.contract import Contract
from platon_aide.delegate import Delegate
from platon_aide.economic import Economic
from platon_aide.govern import Govern
from platon_aide.slashing import Slashing
from platon_aide.staking import Staking
from platon_aide.transfer import Transfer
from platon_aide.graphic import Graphql
from platon_aide.utils import get_web3, send_transaction, ec_recover


class Aide:
    """ 主类，platon各个子模块的集合体，同时支持创建账户、解码等非交易类的操作
    """

    def __init__(self, uri: str, gql_uri: str = None, chain_id: int = None, hrp: str = None):
        self.uri = uri
        self.gql_uri = gql_uri
        self.web3 = get_web3(uri, chain_id, hrp)
        self.web3.middleware_onion.inject(gplaton_poa_middleware, layer=0)
        self.hrp = hrp or self.web3.hrp
        self.chain_id = chain_id or self.web3.platon.chain_id
        # 加入接口和模块
        self.platon = self.web3.platon
        self.admin = self.web3.node.admin
        self.personal = self.web3.node.personal
        self.txpool = self.web3.node.txpool
        self.debug = self.web3.debug
        self.economic = Economic(self.web3)
        self.transfer = Transfer(self.web3)
        self.staking = Staking(self.web3)
        self.delegate = Delegate(self.web3)
        self.slashing = Slashing(self.web3)
        self.govern = Govern(self.web3)
        self.contract = Contract(self.web3)
        self.graphql = Graphql(self.gql_uri, self.web3) if self.gql_uri else None

    def set_default_account(self, account):
        """ 设置默认账户
        """
        self.transfer.set_default_account(account)
        self.staking.set_default_account(account)
        self.delegate.set_default_account(account)
        self.slashing.set_default_account(account)
        self.govern.set_default_account(account)
        self.contract.set_default_account(account)

    def set_result_type(self, result_type):
        self.transfer.set_result_type(result_type)
        self.staking.set_result_type(result_type)
        self.delegate.set_result_type(result_type)
        self.slashing.set_result_type(result_type)
        self.govern.set_result_type(result_type)
        self.contract.set_result_type(result_type)

    def create_account(self):
        """ 创建账户
        """
        account = self.platon.account.create(hrp=self.hrp)
        address = account.address
        private_key = account.privateKey.hex()[2:]
        return address, private_key

    def create_hd_account(self):
        """ 创建HD账户
        """
        # todo: coding
        pass

    def wait_block(self, to_block=None, interval=3):
        """ 等待块高
        """
        current_block = self.platon.block_number
        while current_block < to_block:
            time.sleep(interval)
            current_block = self.platon.block_number

    def ec_recover(self, block_identifier):
        """ 获取出块节点公钥
        """
        block = self.web3.platon.get_block(block_identifier)
        return ec_recover(block)
